/*
 * Copyright (c) 2024 by InvenSense, Inc.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/sys/printk.h>
#include <zephyr/drivers/i2c.h>
#include <zephyr/drivers/gpio.h>

#include <stdint.h>
#include </invn/soniclib/chirp_bsp.h>

// Chirp SonicLib API definitions
#include "chbsp_chirp.h"
#include </invn/soniclib/soniclib.h>

//board config
#include </chirp_board_config.h>

 #include <zephyr/logging/log.h>
//  #define LOG_MODULE CH101_sensor
LOG_MODULE_REGISTER(chbsp_chirp);

#define I2C_INST2  DT_NODELABEL(i2c1)
const struct device *const i2c_ch = DEVICE_DT_GET(I2C_INST2);

const struct device *const ch101_gpio1 = DEVICE_DT_GET(DT_NODELABEL(gpio1));
const struct gpio_dt_spec ch101_int = GPIO_DT_SPEC_GET_OR(CH101_INT, gpios, {0});


#define CH101_PROG_ADDR 0x45;
// uint16_t CH101_APP_ADDR = 0x29;

#define I2C_BUS_CH101   0;

void i2c_bus_test(const struct device *dev, int i2cbus,uint16_t addr)
{
	// for (uint8_t addr = 0; addr < 128; addr++)
	// {
		struct i2c_msg msg = {
				.buf = NULL,
				.len = 0,
				.flags = I2C_MSG_WRITE};

		int ret = i2c_transfer(dev, &msg, 1, addr);
		if (ret == 0)
			LOG_INF("|> i2c_transfer \e[31m0x%02X\e[0m\n", addr);
	// }
}

int chbsp_i2c_init(void){
    int ret = 0;
    return ret;
}

// void test_t(void)
// {
//     // uint16_t addr = 0x45;
//     // uint8_t reg_addr = 0x42 | 0x80;
//     // uint8_t data = 0x40;
// 	// uint8_t buf[] = {reg_addr, (uint8_t)data, (uint8_t)(data >> 8)};

//     // struct i2c_msg msg = {
// 	// 			.buf = (uint8_t) buf,
// 	// 			.len = sizeof(buf),
// 	// 			.flags = I2C_MSG_WRITE};

		// int ret = i2c_transfer(dev, &msg, 1, addr);
// 	// 	if (ret == 0)
// 	// 		// LOG_INF("|> i2c_transfer test\e[31m0x%02X\e[0m\n", addr);

// }

int chbsp_i2c_write(ch_dev_t *dev_ptr, const uint8_t *data, uint16_t num_bytes)
{
    // LOG_INF("chbsp_i2c_write called %d nbytes\n",num_bytes);
    int ret=0;	
    uint8_t address = dev_ptr->i2c_address;
    uint8_t txbuf[num_bytes];//test size
    for(uint8_t i = 0; i < num_bytes; i++) {
        txbuf[i] = data[i];
		// printf("msg[%d]:0x%x\n",i,txbuf[i]);
	}   
  
    /*i2c_transfer call*/
    struct i2c_msg msg[1];
				msg[0].buf = data,
				msg[0].len = num_bytes,
				msg[0].flags = I2C_MSG_WRITE | I2C_MSG_STOP;     
    ret = i2c_transfer(i2c_ch, &msg[0], 1, (uint16_t)address);
        
    // ret = i2c_write(i2c_ch, txbuf, num_bytes, (uint16_t)address);
    if (ret != 0) {
		LOG_ERR("chbsp_i2c_write %dbytes failed to addr:0x%x err:%d\n",num_bytes, address, ret);        
	}
    else
    // LOG_INF("chbsp_i2c_write %dbytes success to addr:0x%x\n",num_bytes, address);
    	       	
	return ret;
}

int chbsp_i2c_read(ch_dev_t *dev_ptr, uint8_t *data, uint16_t num_bytes)
{   
    int ret;
    uint8_t address = dev_ptr->i2c_address;
    struct i2c_msg msg[1];
    msg[0].buf = data;
    msg[0].len = num_bytes;
    msg[0].flags = I2C_MSG_READ | I2C_MSG_STOP;
    // LOG_INF("chbsp_i2c_read called\n");

    // ret = i2c_read(i2c_ch, msg[0], 1,(uint16_t)address);
    ret = i2c_transfer(i2c_ch, &msg[0], 1, (uint16_t)address);
    if(ret != 0){
	LOG_ERR("chbsp_i2c_read from addr:0x%x failed: %d\n",address, ret);    
    }
    else{
    //     for(uint8_t i = 0; i < num_bytes; i++)  
	// 	printf("data[%d]:0x%x\n",i,data[i]); 
    //    LOG_INF("chbsp_i2c_read success from addr:0x%x\n",address);
    }

    return ret;
}

int chbsp_i2c_mem_write(ch_dev_t *dev_ptr, uint16_t mem_addr, uint8_t *data, uint16_t num_bytes)
{     
    int ret=0;
    uint8_t i2c_addr = ch_get_i2c_address(dev_ptr);
    LOG_INF("chbsp_i2c_mem_write mem_addr:0x%x i2c_addr:0x%x\n",mem_addr,i2c_addr);   	
    
    uint8_t txbuf[num_bytes];
    // txbuf[0] = mem_addr;
    // LOG_INF("add[%d]:0x%x",i,txbuf[0]);
    // for(uint8_t i = 0; i < num_bytes; i++) {
    //     txbuf[i+1] = data[i];
	// 	LOG_INF("msg[%d]:0x%x",i,txbuf[i+1]);
	// }        

    struct i2c_msg msgs[2];
    uint8_t reg = mem_addr;
    msgs[0].buf = reg;
    msgs[0].len = 1;
    msgs[0].flags = I2C_MSG_WRITE;
    msgs[1].buf = data;
    msgs[1].len = num_bytes;
    msgs[1].flags = I2C_MSG_WRITE | I2C_MSG_STOP;
    ret = i2c_transfer(i2c_ch, msgs, 2,(uint16_t)i2c_addr);

    // ret = i2c_write(i2c_ch, txbuf, num_bytes, (uint16_t)address);
    if (ret != 0) {
		LOG_ERR("chbsp_i2c_mem_write %dbytes failed to addr:0x%x err:%d\n",num_bytes, i2c_addr, ret);          
	}
    else
    LOG_INF("chbsp_i2c_mem_write success!/n");

    return ret; 
}


int chbsp_i2c_mem_read(ch_dev_t *dev_ptr, uint16_t mem_addr, uint8_t *data, uint16_t num_bytes)
{   
    int ret=0;    
    uint8_t i2c_addr = ch_get_i2c_address(dev_ptr);
    LOG_INF("chbsp_i2c_mem_read i2c_addr:0x%x mem_addr:0x%x\n",i2c_addr,mem_addr);    
    
    // struct i2c_msg msgs[2];
    // uint8_t reg = mem_addr;
    // msgs[0].buf = reg;
    // msgs[0].len = 1;
    // msgs[0].flags = I2C_MSG_WRITE;
    // msgs[1].buf = (uint8_t *)data;
    // msgs[1].len = num_bytes;
    // msgs[1].flags = I2C_MSG_RESTART | I2C_MSG_READ | I2C_MSG_STOP;
    // ret = i2c_transfer(i2c_ch, msgs, 2,(uint16_t)i2c_addr);

    //     uint8_t writebuf[1];
    //   writebuf[0] = (uint8_t)mem_addr;  
    //   ret =  i2c_write_read(i2c_ch, (uint16_t)i2c_addr, writebuf[0], 1, &data, num_bytes);

    ret = i2c_burst_read(i2c_ch,(uint16_t)i2c_addr, (uint8_t)mem_addr, data, num_bytes);
    if (ret != 0) {
		LOG_ERR("chbsp_i2c_mem_read %dbytes failed from addr:0x%x err:%d\n",num_bytes, mem_addr, ret);          
	}
    else{
    for(uint8_t i = 0; i < num_bytes; i++)  
		LOG_INF("data[%d]:0x%02X",i,data[i]);
        LOG_INF("chbsp_i2c_mem_read %dbytes success from addr:0x%x\n",num_bytes, mem_addr);
     
    }
    return ret;
}

uint8_t chbsp_i2c_get_info(ch_group_t* grp_ptr, uint8_t io_index, ch_i2c_info_t *info_ptr)
{  
//    return i2c_check();
    if (!device_is_ready(i2c_ch)) {
		LOG_ERR("i2c_ch101 is not ready!\n");
         		return 1;
	}
    else 
    // LOG_INF("i2c_ch101 is ready!\n");      
    {
       uint8_t ret_val = 1;
	    if (io_index <= CHIRP_MAX_NUM_SENSORS) {
		info_ptr->address = 45;//CH101_PROG_ADDR;//=0x45
		info_ptr->bus_num = I2C_BUS_CH101;//=1
		info_ptr->drv_flags = 0;	// no special I2C handling by SonicLib driver is needed
		ret_val = 0;
	    }
	    return ret_val;
    }
}

void chbsp_program_disable(ch_dev_t *dev_ptr)
{
    int err=0;
    // LOG_INF("chbsp_program_disable \n");
     err = gpio_pin_set(ch101_gpio1,8, 0);
        if (err) {
            LOG_ERR("Cannot write Prog gpio");
        }
}

void chbsp_program_enable(ch_dev_t *dev_ptr)
{
     int err=0;
    //  LOG_INF("chbsp_program_enable \n");
     err = gpio_pin_set(ch101_gpio1,8, 1);
        if (err) {
            LOG_ERR("Cannot write Prog gpio\n");
        }
}

void chbsp_delay_ms(uint32_t ms){

    // LOG_INF("chbsp_delay_ms: %d\n",ms);
    k_msleep (ms);
}

void chbsp_reset_release(void)
{
    int err=0;
    //  LOG_INF("chbsp_reset_release \n");
     err = gpio_pin_set(ch101_gpio1,9, 1);
        if (err) {
            LOG_ERR("Cannot write Rest gpio");
        }
}

void chbsp_reset_assert(void)
{
    //  LOG_INF("chbsp_reset_assert \n");
     int err = gpio_pin_set(ch101_gpio1,9, 0);
      if (err) {
            LOG_ERR("Cannot write Rest gpio");
        }
}

void chbsp_set_int1_dir_out(ch_dev_t *dev_ptr)
{
    //INT as input to host
    gpio_pin_set(ch101_gpio1,7, 1);
}

void chbsp_set_int1_dir_in(ch_dev_t *dev_ptr)
{
    //INT as output
    gpio_pin_set(ch101_gpio1,7, 0);
}

void chbsp_int1_interrupt_enable(ch_dev_t *dev_ptr)
{
    gpio_pin_interrupt_configure_dt(&ch101_int,GPIO_INT_EDGE_TO_ACTIVE);
}

void chbsp_int1_interrupt_disable(ch_dev_t *dev_ptr)
{
    gpio_pin_interrupt_configure_dt(&ch101_int,GPIO_INT_DISABLE);
}

void chbsp_int1_clear(ch_dev_t *dev_ptr)
{    
        LOG_INF("INT clear\n");        
        gpio_pin_set(ch101_gpio1,6, 0);
}

void chbsp_int1_set(ch_dev_t *dev_ptr)
{
        LOG_INF("INT set\n");              
        gpio_pin_set(ch101_gpio1,6, 1);
}

void chbsp_group_int1_clear(ch_group_t *grp_ptr)
{
    LOG_INF("gpr INT clear\n");
    gpio_pin_set(ch101_gpio1,6, 0);
}

void chbsp_group_int1_set(ch_group_t *grp_ptr)
{
    LOG_INF("gpr INT set\n");
    gpio_pin_set(ch101_gpio1,6, 1);
}

uint32_t chbsp_timestamp_ms(void)
{
    //LOG_INF("chbsp_timestamp_ms called\n");
    return k_uptime_get_32();
}

void chbsp_i2c_reset(ch_dev_t *dev_ptr)
{
    // LOG_INF("chbsp_i2c_reset \n");
    int result = i2c_configure(i2c_ch, I2C_SPEED_SET(I2C_SPEED_FAST));
     if (result) {
            LOG_ERR("Cannot set I2C_SPEED_FAST\n");
        }
}

void chbsp_delay_us(uint32_t us)
{
    LOG_INF("chbsp_delay_us: %d\n",us);
    k_usleep (us);
}

/*-----------ported code from .cpp to c---make changes to access nrf i2C API----------*/
/*

CHx01* sensor_group_ptr = NULL;
static int rst_pin_id;
static int reset_n;
static void int1_0_handler(void);
static void int1_1_handler(void);
const chx01_dev_irq_handler *irq_handlers[2] = {int1_0_handler, int1_1_handler};

int chbsp_i2c_init(void)
{
    // Initialize the I2C interface
    Wire.begin();
    Wire.setClock(DEFAULT_I2C_CLOCK);
    return 0;
}

void chbsp_module_init(int rst_id, bool rst_n)
{
    // Store reset pin ID and reset polarity
    rst_pin_id = rst_id;
    reset_n = rst_n;
}

void chbsp_i2c_reset(ch_dev_t *dev_ptr)
{
    // Cast device pointer and call its I2C reset function
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    device->i2c_reset(device);
}

static void int1_0_handler(void)
{
    if(sensor_group_ptr != NULL)
    {
        // Handle interrupt for device 0
        ch_interrupt(sensor_group_ptr, 0);
        // Enable interrupt again after handling
        chbsp_int1_interrupt_enable(ch_get_dev_ptr(sensor_group_ptr, 0));
    }
}

static void int1_1_handler(void)
{
    if(sensor_group_ptr != NULL)
    {
        // Handle interrupt for device 1
        ch_interrupt(sensor_group_ptr, 1);
        // Enable interrupt again after handling
        chbsp_int1_interrupt_enable(ch_get_dev_ptr(sensor_group_ptr, 1));
    }
}

void board_init(CHx01 *grp_ptr)
{
    // Make local copy of group pointer
    sensor_group_ptr = grp_ptr;

    // Configure Reset pin as output, set to 1 (Reset enabled, inverted to RST_N on the board)
    pinMode(rst_pin_id, OUTPUT);
    digitalWrite(rst_pin_id, reset_n ? LOW : HIGH);
}

void chbsp_set_int1_dir_out(ch_dev_t *dev_ptr)
{
    // Set interrupt pin direction to output
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    device->set_int1_dir(device, CHX01_INT_DIR_OUT);
}

void chbsp_set_int1_dir_in(ch_dev_t *dev_ptr)
{
    // Set interrupt pin direction to input
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    device->set_int1_dir(device, CHX01_INT_DIR_IN);
}

void chbsp_int1_clear(ch_dev_t *dev_ptr)
{
    // Clear interrupt pin
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    device->set_int1(device, LOW);
}

void chbsp_int1_set(ch_dev_t *dev_ptr)
{
    // Set interrupt pin
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    device->set_int1(device, HIGH);
}

void chbsp_int1_interrupt_enable(ch_dev_t *dev_ptr)
{
    // Enable interrupt for the device
    CHx01_dev* device = (CHx01_dev*)dev_ptr;

    if(device->io_index < CHIRP_MAX_NUM_SENSORS)
    {
        device->enableInterrupt(device, irq_handlers[device->io_index]);
    }
}

void chbsp_int1_interrupt_disable(ch_dev_t *dev_ptr)
{
    // Disable interrupt for the device
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    device->disableInterrupt(device);
}

uint8_t chbsp_i2c_get_info(ch_group_t* grp_ptr, uint8_t io_index, ch_i2c_info_t *info_ptr)
{
    // Get I2C information for the device at the specified index
    CHx01* group_ptr = (CHx01*)grp_ptr;
    CHx01_dev* device = group_ptr->get_device(group_ptr, io_index);
    if(device != NULL)
    {
        info_ptr->address   = device->get_i2c_addr(device);
        // All sensors on the same bus
        info_ptr->bus_num   = 0;
        info_ptr->drv_flags = 0; // No special I2C handling by SonicLib driver is needed
        return 0;
    }
    return -1;
}

int chbsp_i2c_write(ch_dev_t *dev_ptr, const uint8_t *data, uint16_t num_bytes)
{
    // Write data to the I2C bus
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    TwoWire* i2c = device->get_i2c(device);
    uint16_t offset = 0;

    while(offset < num_bytes)
    {
        i2c->beginTransmission(ch_get_i2c_address(dev_ptr));
        uint16_t length = ((num_bytes - offset) > ARDUINO_I2C_BUFFER_LENGTH) ? ARDUINO_I2C_BUFFER_LENGTH : (num_bytes - offset);
        i2c->write(&data[offset], length);
        offset += length;
        i2c->endTransmission((offset == num_bytes));
    }

    return 0;
}

int chbsp_i2c_read(ch_dev_t *dev_ptr, uint8_t *data, uint16_t num_bytes)
{
    // Read data from the I2C bus
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    TwoWire* i2c = device->get_i2c(device);
    uint16_t offset = 0;

    while(offset < num_bytes)
    {
        uint16_t rx_bytes = 0;
        i2c->beginTransmission(ch_get_i2c_address(dev_ptr));
        uint16_t length = ((num_bytes - offset) > ARDUINO_I2C_BUFFER_LENGTH) ? ARDUINO_I2C_BUFFER_LENGTH : (num_bytes - offset);
        rx_bytes = i2c->requestFrom((int)ch_get_i2c_address(dev_ptr), length);
        if (rx_bytes == length) {
            for(uint8_t i = 0; i < length; i++) {
                data[offset+i] = i2c->read();
            }
            offset += length;
            i2c->endTransmission((offset == num_bytes));
        } else {
            i2c->endTransmission((offset == num_bytes));
        }
    }
    if(offset == num_bytes)
    {
        return 0;
    } else {
        return -1;
    }
}

int chbsp_i2c_mem_write(ch_dev_t *dev_ptr, uint16_t mem_addr, uint8_t *data, uint16_t num_bytes)
{
    // Write data to a specific memory address on the I2C bus
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    TwoWire* i2c = device->get_i2c(device);

    i2c->beginTransmission(ch_get_i2c_address(dev_ptr));
    i2c->write((uint8_t)mem_addr);
    // TODO: the doc says we should send num_bytes here
    for(uint8_t i = 0; i < num_bytes; i++) {
        i2c->write(data[i]);
    }
    i2c->endTransmission();
    return 0;
}

int chbsp_i2c_mem_read(ch_dev_t *dev_ptr, uint16_t mem_addr, uint8_t *data, uint16_t num_bytes)
{
    // Read data from a specific memory address on the I2C bus
    CHx01_dev* device = (CHx01_dev*)dev_ptr;
    TwoWire* i2c = device->get_i2c(device);
    uint16_t rx_bytes = 0;
    i2c->beginTransmission(ch_get_i2c_address(dev_ptr));
    i2c->write((uint8_t)mem_addr);
    // TODO: the doc says we should send num_bytes here
    i2c->endTransmission(false);

    i2c->beginTransmission(ch_get_i2c_address(dev_ptr));
    rx_bytes = i2c->requestFrom((int)ch_get_i2c_address(dev_ptr), num_bytes);
    if (rx_bytes == num_bytes) {
        for(uint8_t i = 0; i < num_bytes; i++) {
            data[i] = i2c->read();
        }
        i2c->endTransmission();
        return 0;
    } else {
        return -1;
    }
}

void chbsp_reset_assert(void)
{
    // Assert the reset pin
    digitalWrite(rst_pin_id, reset_n ? LOW : HIGH);
}

void chbsp_reset_release(void)
{
    // Release the reset pin
    digitalWrite(rst_pin_id, reset_n ? HIGH : LOW);
}

void chbsp_program_enable(ch_dev_t *dev_ptr)
{
    // Enable programming mode
    ((CHx01_dev*)dev_ptr)->set_prog((CHx01_dev*)dev_ptr, HIGH);
}

void chbsp_program_disable(ch_dev_t *dev_ptr)
{
    // Disable programming mode
    ((CHx01_dev*)dev_ptr)->set_prog((CHx01_dev*)dev_ptr, LOW);
}

void chbsp_delay_us(uint32_t us)
{
    // Delay for the specified number of microseconds
    delayMicroseconds(us);
}

void chbsp_delay_ms(uint32_t ms)
{
    // Delay for the specified number of milliseconds
    delay(ms);
}

void chbsp_group_set_int1_dir_out(ch_group_t *grp_ptr)
{
    // Set interrupt pin direction to output for all devices in the group
    for(int i = 0; i < grp_ptr->num_ports; i++)
    {
        chbsp_set_int1_dir_out(ch_get_dev_ptr(grp_ptr, i));
    }
}

void chbsp_group_set_int1_dir_in(ch_group_t *grp_ptr)
{
    // Set interrupt pin direction to input for all devices in the group
    for(int i = 0; i < grp_ptr->num_ports; i++)
    {
        chbsp_set_int1_dir_in(ch_get_dev_ptr(grp_ptr, i));
    }
}

void chbsp_group_int1_clear(ch_group_t *grp_ptr)
{
    // Clear interrupt pin for all devices in the group
    for(int i = 0; i < grp_ptr->num_ports; i++)
    {
        chbsp_int1_clear(ch_get_dev_ptr(grp_ptr, i));
    }
}

void chbsp_group_int1_set(ch_group_t *grp_ptr)
{
    // Set interrupt pin for all devices in the group
    for(int i = 0; i < grp_ptr->num_ports; i++)
    {
        chbsp_int1_set(ch_get_dev_ptr(grp_ptr, i));
    }
}

void chbsp_group_int1_interrupt_enable(ch_group_t *grp_ptr)
{
    // Enable interrupt for all devices in the group
    for(int i = 0; i < grp_ptr->num_ports; i++)
    {
        chbsp_int1_interrupt_enable(ch_get_dev_ptr(grp_ptr, i));
    }
}

void chbsp_group_int1_interrupt_disable(ch_group_t *grp_ptr)
{
    // Disable interrupt for all devices in the group
    for(int i = 0; i < grp_ptr->num_ports; i++)
    {
        chbsp_int1_interrupt_disable(ch_get_dev_ptr(grp_ptr, i));
    }
}

uint32_t chbsp_timestamp_ms(void)
{
    // Return the current timestamp in milliseconds
    return (uint32_t) millis();
}
*/


